/*
*  Power BI Visual CLI
*
*  Copyright (c) Microsoft Corporation
*  All rights reserved.
*  MIT License
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the ""Software""), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
*  THE SOFTWARE.
*/
"use strict";

import "core-js/stable";
import "./../style/visual.less";
import "./css/picsScroller.css";
import powerbi from "powerbi-visuals-api";
import VisualConstructorOptions = powerbi.extensibility.visual.VisualConstructorOptions;
import VisualUpdateOptions = powerbi.extensibility.visual.VisualUpdateOptions;
import IVisual = powerbi.extensibility.visual.IVisual;
import EnumerateVisualObjectInstancesOptions = powerbi.EnumerateVisualObjectInstancesOptions;
import VisualObjectInstance = powerbi.VisualObjectInstance;
import DataView = powerbi.DataView;
import VisualObjectInstanceEnumerationObject = powerbi.VisualObjectInstanceEnumerationObject;
import { VisualSettings, TagSettings ,TagPosition } from "./settings";
import * as d3 from "d3";
import {pixelConverter} from "powerbi-visuals-utils-typeutils";

export interface IPicsScrollerData {
    urls: string[];
    tags?:string[];
}
export class Visual implements IVisual {
    private container: HTMLElement;
    private settings: VisualSettings;
    private host:powerbi.extensibility.visual.IVisualHost;
    private selectionManager: powerbi.extensibility.ISelectionManager;
    private innerContainer!: HTMLDivElement;
    private animationPlaying: boolean = true;//控制动画暂停
    private data:IPicsScrollerData;
    private containerSize:number[];
    constructor(options: VisualConstructorOptions) {
        this.container = options.element;
        this.host=options.host;
        this.selectionManager=this.host.createSelectionManager();
        d3.select(this.container).on("contextmenu", () => {
            const mouseEvent: MouseEvent = <MouseEvent>d3.event;
            this.selectionManager.showContextMenu({}, {
                x: mouseEvent.clientX,
                y: mouseEvent.clientY
            });
            mouseEvent.preventDefault();
        });
    }
    private initial() {
        if (this.innerContainer) {
            this.innerContainer.remove();
        };
        this.innerContainer = document.createElement("div");
        this.innerContainer.className = "pics-sroller-inner-container";
        this.container.appendChild(this.innerContainer);
    }
    public update(options: VisualUpdateOptions) {

        let settings = Visual.parseSettings(options && options.dataViews && options.dataViews[0]);
        let urls:string[]=[];
        let tags:string[]=[];
        if(options.dataViews&&options.dataViews[0]&&options.dataViews[0].categorical){
            let categorical=options.dataViews[0].categorical;
            if(categorical.categories&&categorical.categories.length>0){
                categorical.categories.map((field,index,array)=>{
                    if(field&&field.values&&field.values.length>0){
                        if(field.source.roles["urls"]){
                            urls=<string[]>field.values;
                        }else if(field.source.roles["tags"]){
                            tags=<string[]>field.values;
                        }
                    }
                });
            }
        }else{
            this.innerContainer.remove();
            return;
        }

        let data: IPicsScrollerData = { urls:urls ,tags:tags};
        let containerSize=[options.viewport.width,options.viewport.height];

        //处理输入
        //feet
        settings.animation.stepLength=(settings.animation.stepLength>=0&&settings.animation.stepLength<=100)?settings.animation.stepLength:1;
        let feet = options.viewport.width / 100 * settings.animation.stepLength;
        //interval
        settings.animation.interval=settings.animation.interval>0?settings.animation.interval:100;
        let interval = settings.animation.interval;
        //rowGap
        settings.layout.rowGap=(settings.layout.rowGap>=0&&settings.layout.rowGap<=100)?settings.layout.rowGap:0;
        let rowGap = options.viewport.height / 100 * settings.layout.rowGap;
        //columnGap
        settings.layout.columnGap=(settings.layout.columnGap>=0&&settings.layout.columnGap<=100)?settings.layout.columnGap:0;
        let columnGap = options.viewport.width / 100 * settings.layout.columnGap;
        //rowsCount
        settings.layout.rowsCount=settings.layout.rowsCount>=1?Math.floor(settings.layout.rowsCount):1;
        let rowsCount=settings.layout.rowsCount;

        //tag
        let tagFontSize:number=12;
        if(settings.tag.show){
            tagFontSize=pixelConverter.fromPointToPixel(settings.tag.textSize);
        }

        if (!(data.urls.length > 0) || !(rowsCount > 0)) {
            console.info("picsScroller error: (source_url.length,rowsCount) ", data.urls.length + "_" + rowsCount);
            return;
        };
        if (rowsCount > data.urls.length) {
            rowsCount = data.urls.length;
        };

        let justifyEqual=function(a:any,b:any){
            return JSON.stringify(a)==JSON.stringify(b);
        };
        //判断是否完全无需刷新
        if(justifyEqual(this.data,data)&&justifyEqual(this.settings,settings)&&justifyEqual(this.containerSize,containerSize)){
            return;
        }

        //必要的类
        this.initial();
        this.data=data;
        this.settings=settings;
        this.containerSize=containerSize;

        //数据转换
        let brandsCount = data.urls.length;
        let unitsCountPerRow = Math.floor(brandsCount / rowsCount);
        let logoUrlsArr: string[][] = [];
        let tagsArr:string[][]=[];
        //转换url列表形态
        for (let i = 0; i < rowsCount; i++) {
            logoUrlsArr.push(data.urls.slice(i * unitsCountPerRow, Math.min((i + 1) * unitsCountPerRow, data.urls.length)));
            if(data.tags&&data.tags.length>0){
                tagsArr.push(data.tags.slice(i * unitsCountPerRow, Math.min((i + 1) * unitsCountPerRow, data.tags.length)));
            };
        };
        //分成rowsCount行
        for (let rowNum = 0; rowNum < rowsCount; rowNum++) {
            let rowContainer = document.createElement('div');
            this.innerContainer.appendChild(rowContainer);
            rowContainer.classList.add('pics-scroller-row-container');

            rowContainer.setAttribute('id', 'pics-scroller-row-container' + rowNum);
            rowContainer.setAttribute("style", "height:" + 100 / rowsCount + "%;");
            rowContainer.setAttribute("data-animation-playing", 'false');
            rowContainer.setAttribute("data-display-placehold", "false");
            //Add rowGap
            if (rowNum > 0) {
                rowContainer.style.marginTop = rowGap + "px";
            };
            //双倍图片容器，第一组用来显示，第二组放上去暂时隐藏
            //如果第一组发生了溢出，就显示出第二组并应用滚动动画
            for (let i = 0; i < 2; i++) {
                for (let j = 0; j < logoUrlsArr[rowNum].length; j++) {
                    let logoContainer = document.createElement('div');
                    logoContainer.className = "pics-scroller-img-container";
                    if (i == 1) {
                        logoContainer.classList.add("pics-scroll-invisible-img-container");
                    };
                    let logo = document.createElement('img');
                    logo.style.margin = "0px " + columnGap / 2 + "px";
                    logo.setAttribute("class", "pics-scroller-img");
                    logo.setAttribute("src", logoUrlsArr[rowNum][j]);
                    logoContainer.appendChild(logo);
                    rowContainer.appendChild(logoContainer);
                    //tag patch
                    if(settings.tag.show&&tagsArr&&tagsArr.length>0){
                        logo.style.height="calc(100% - "+1.5*tagFontSize+"px)";
                        let tagElement=document.createElement("p");
                        tagElement.innerText=tagsArr[rowNum][j];
                        tagElement.className="pics-scroller-tag";
                        tagElement.style.fontSize=tagFontSize+"px";
                        tagElement.style.height=1.5*tagFontSize+"px";
                        tagElement.style.color=settings.tag.fontColor.toString();
                        tagElement.style.opacity=(100-settings.tag.transparency)+"%";
                        if(settings.tag.position.toString()==TagPosition[TagPosition.above].toString()){
                            logoContainer.insertBefore(tagElement,logo);
                        }else{
                            logoContainer.appendChild(tagElement);
                        }
                    }
                };
            }

            //重新激活动画
            if (this.animationPlaying) {
                this.activateAnimation();
            }
            //假无限滚动
            window.setInterval(this.scrollLeftInfinity, interval, rowContainer, feet);
        };

    }
    public activateAnimation() {
        this.animationPlaying = true;
        let rowContainers = this.innerContainer.getElementsByClassName("pics-scroller-row-container");
        for (let i = 0; i < rowContainers.length; i++) {
            rowContainers[i].setAttribute("data-animation-playing", "true");
        }
    }
    private suspendAnimation() {
        let rowContainers = this.innerContainer.getElementsByClassName("pics-scroller-row-container");
        for (let i = 0; i < rowContainers.length; i++) {
            rowContainers[i].setAttribute("data-animation-playing", "false");
        }
    }
    //定义无限向左滚动函数
    private scrollLeftInfinity(obj: HTMLDivElement, feet: number) {
        let playAnimation = <string>obj.getAttribute("data-animation-playing");
        let displayPlaceHold = <string>obj.getAttribute("data-display-placehold");
        if (playAnimation == "false") {
            return;
        } else if (playAnimation == "true" && displayPlaceHold == "false" && obj.scrollWidth > obj.clientWidth) {
            let invisibleImgContainers = obj.getElementsByClassName("pics-scroll-invisible-img-container");
            while (invisibleImgContainers.length > 0) {
                invisibleImgContainers[0].classList.remove("pics-scroll-invisible-img-container");
            }
            obj.setAttribute("data-display-placehold", "true");
        }
        if (obj.scrollWidth > obj.clientWidth) {
            if (obj.scrollLeft >= obj.scrollWidth / 2) {
                obj.scrollLeft = 0;
            } else {
                obj.scrollLeft += feet;
            };
        }

    }


    private static parseSettings(dataView: DataView): VisualSettings {
        return <VisualSettings>VisualSettings.parse(dataView);
    }

    /**
     * This function gets called for each of the objects defined in the capabilities files and allows you to select which of the
     * objects and properties you want to expose to the users in the property pane.
     *
     */
    public enumerateObjectInstances(options: EnumerateVisualObjectInstancesOptions): VisualObjectInstance[] | VisualObjectInstanceEnumerationObject {
        return VisualSettings.enumerateObjectInstances(this.settings || VisualSettings.getDefault(), options);
    }
}